
#include <daemon/adafs_ops/metadentry.hpp>
#include <daemon/backend/metadata/db.hpp>
#include <daemon/backend/data/chunk_storage.hpp>

using namespace std;

ino_t generate_inode_no() {
    std::lock_guard<std::mutex> inode_lock(ADAFS_DATA->inode_mutex);
    // TODO check that our inode counter is within boundaries of inode numbers in the given node
    return ADAFS_DATA->raise_inode_count(1);
}

/**
 * Creates metadata (if required) and dentry at the same time
 * @param path
 * @param mode
 */
void create_metadentry(const std::string& path, mode_t mode) {

    Metadata md{mode};
    // update metadata object based on what metadata is needed
    if (ADAFS_DATA->atime_state() || ADAFS_DATA->mtime_state() || ADAFS_DATA->ctime_state()) {
        std::time_t time;
        std::time(&time);
        auto time_s = std::to_string(time);
        if (ADAFS_DATA->atime_state())
            md.atime(time);
        if (ADAFS_DATA->mtime_state())
            md.mtime(time);
        if (ADAFS_DATA->ctime_state())
            md.ctime(time);
    }
    if (ADAFS_DATA->uid_state())
        md.uid(getuid());
    if (ADAFS_DATA->gid_state())
        md.gid(getgid());
    if (ADAFS_DATA->inode_no_state())
        md.inode_no(generate_inode_no());

    ADAFS_DATA->mdb()->put(path, md.serialize());
}

std::string get_metadentry_str(const std::string& path) {
        return ADAFS_DATA->mdb()->get(path);
}

/**
 * Returns the metadata of an object at a specific path. The metadata can be of dummy values if configured
 * @param path
 * @param attr
 * @return
 */
Metadata get_metadentry(const std::string& path) {
    return {get_metadentry_str(path)};
}

/**
 * Remove metadentry if exists and try to remove all chunks for path
 * @param path
 * @return
 */
void remove_node(const string& path) {
    ADAFS_DATA->mdb()->remove(path); // remove metadentry
    ADAFS_DATA->storage()->destroy_chunk_space(path); // destroys all chunks for the path on this node
}

/**
 * Gets the size of a metadentry
 * @param path
 * @param ret_size (return val)
 * @return err
 */
size_t get_metadentry_size(const string& path) {
    return get_metadentry(path).size();
}

/**
 * Updates a metadentry's size atomically and returns the corresponding size after update
 * @param path
 * @param io_size
 * @return the updated size
 */
void update_metadentry_size(const string& path, size_t io_size, off64_t offset, bool append) {
#ifdef LOG_TRACE
    ADAFS_DATA->mdb()->iterate_all();
#endif
    ADAFS_DATA->mdb()->increase_size(path, io_size + offset, append);
#ifdef LOG_TRACE
    ADAFS_DATA->mdb()->iterate_all();
#endif
}

void update_metadentry(const string& path, Metadata& md) {
    ADAFS_DATA->mdb()->update(path, path, md.serialize());
}

/**
 * @param path of the object whose permissions are checked
 * @param mask single bit mask to check against
 * @return errno
 */
int check_access_mask(const string& path, const int mask) {
    Metadata md;
    try {
        md = get_metadentry(path);
    } catch (const NotFoundException& e) {
        return ENOENT;
    }

    /*
     * if only check if file exists is wanted, return success.
     * According to POSIX (access(2)) the mask is either the value F_OK,
     * or a mask consisting of the bitwise OR of one or more of R_OK, W_OK, and X_OK.
     */
    if (mask == F_OK)
        return 0;

    // root user is a god
    if (ADAFS_DATA->uid_state() && md.uid() == 0)
        return 0;

    // We do not check for the actual user here, because the cluster batchsystem should take care of it
    // check user leftmost 3 bits for rwx in md->mode
    if (ADAFS_DATA->uid_state()) {
        // Because mode comes only with the first 3 bits used, the user bits have to be shifted to the right to compare
        if ((mask & md.mode() >> 6) == static_cast<unsigned int>(mask))
            return 0;
        else
            return EACCES;
    }

    // check group middle 3 bits for rwx in md->mode
    if (ADAFS_DATA->gid_state()) {
        if ((mask & md.mode() >> 3) == static_cast<unsigned int>(mask))
            return 0;
        else
            return EACCES;
    }

    // check other rightmost 3 bits for rwx in md->mode.
    // Because they are the rightmost bits they don't need to be shifted
    if ((mask & md.mode()) == static_cast<unsigned int>(mask)) {
        return 0;
    }

    return EACCES;
}

std::vector<std::pair<std::string, bool>> get_dirents(const std::string& dir){
    return ADAFS_DATA->mdb()->get_dirents(dir);
}